/* 
**  Xbox360 USB Gamepad Userspace Driver
**  Copyright (C) 2008 Ingo Ruhnke <grumbel@gmx.de>
**
**  This program is free software: you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation, either version 3 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <iostream>
#include <stdexcept>
#include <boost/bind.hpp>
#include <boost/format.hpp>

#include "arg_parser.hpp"
#include "helper.hpp"
#include "command_line_options.hpp"

// See http://stackoverflow.com/questions/303562/c-format-macro-inline-ostringstream
#define RAISE_EXCEPTION(x) throw std::runtime_error(static_cast<std::ostringstream&>(std::ostringstream().seekp(0, std::ios_base::cur) << x).str())

CommandLineOptions* command_line_options = 0;

enum {
  OPTION_HELP,
  OPTION_VERBOSE,
  OPTION_VERSION,
  OPTION_QUIET,
  OPTION_SILENT,
  OPTION_DAEMON,
  OPTION_TEST_RUMBLE,
  OPTION_RUMBLE,
  OPTION_QUIT,
  OPTION_NO_UINPUT,
  OPTION_MIMIC_XPAD,
  OPTION_TYPE,
  OPTION_FORCE_FEEDBACK,
  OPTION_RUMBLE_GAIN,
  OPTION_BUTTONMAP,
  OPTION_AXISMAP,
  OPTION_NAME,
  OPTION_UI_CLEAR,
  OPTION_UI_AXISMAP,
  OPTION_UI_BUTTONMAP,
  OPTION_ID,
  OPTION_WID,
  OPTION_LED,
  OPTION_DPAD_ONLY,
  OPTION_DPAD_AS_BUTTON,
  OPTION_DEADZONE,
  OPTION_DEADZONE_TRIGGER,
  OPTION_TRIGGER_AS_BUTTON,
  OPTION_TRIGGER_AS_ZAXIS,
  OPTION_AUTOFIRE,
  OPTION_CALIBRARIOTION,
  OPTION_RELATIVE_AXIS,
  OPTION_SQUARE_AXIS,
  OPTION_HELP_LED,
  OPTION_DEVICE_BY_ID,
  OPTION_DEVICE_BY_PATH,
  OPTION_LIST_SUPPORTED_DEVICES,
  OPTION_LIST_CONTROLLER,
  OPTION_HELP_DEVICES
};

CommandLineOptions::CommandLineOptions()
{
  mode     = RUN_DEFAULT;

  verbose  = false;
  silent   = false;
  quiet    = false;
  rumble   = false;
  led      = -1;
  rumble_l = -1;
  rumble_r = -1;
  rumble_gain = 255;
  controller_id = 0;
  wireless_id   = 0;
  instant_exit = false;
  no_uinput = false;
  gamepad_type = GAMEPAD_UNKNOWN;
  busid[0] = '\0';
  devid[0] = '\0';
  vendor_id  = -1;
  product_id = -1;
  deadzone = 0;
  deadzone_trigger = 0;
  square_axis  = false;

  argp
    .add_usage("[OPTION]...")
    .add_text("Xbox360 USB Gamepad Userspace Driver")
    .add_newline()
  
    .add_text("General Options: ")
    .add_option(OPTION_HELP,         'h', "help",         "", "display this help and exit")
    .add_option(OPTION_VERSION,      'V', "version",      "", "print the version number and exit")
    .add_option(OPTION_VERBOSE,      'v', "verbose",      "", "print verbose messages")
    .add_option(OPTION_HELP_LED,      0,  "help-led",     "", "list possible values for the led")
    .add_option(OPTION_HELP_DEVICES,  0,  "help-devices", "", "list supported devices")
    .add_option(OPTION_SILENT,       's', "silent",  "",  "do not display events on console")
    .add_option(OPTION_QUIET,         0,  "quiet",   "",  "do not display startup text")
    .add_option(OPTION_ID,           'i', "id",      "N", "use controller with id N (default: 0)")
    .add_option(OPTION_WID,          'w', "wid",     "N", "use wireless controller with wid N (default: 0)")
    .add_option(OPTION_LIST_CONTROLLER, 'L', "list-controller", "", "list available controllers")
    .add_option(OPTION_LIST_SUPPORTED_DEVICES, 0, "list-supported-devices", "", "list supported devices (used by xboxdrv-daemon.py)")
    .add_option(OPTION_TEST_RUMBLE,  'R', "test-rumble", "", "map rumbling to LT and RT (for testing only)")
    .add_option(OPTION_NO_UINPUT,     0,  "no-uinput",   "", "do not try to start uinput event dispatching")
    .add_option(OPTION_MIMIC_XPAD,    0,  "mimic-xpad",  "", "Causes xboxdrv to use the same axis and button names as the xpad kernel driver")
    .add_option(OPTION_DAEMON,       'D', "daemon",      "", "run as daemon")
    .add_newline()

    .add_text("Device Options: ")
    .add_option(OPTION_DEVICE_BY_PATH, 0, "device-by-path", "BUS:DEV", "Use device BUS:DEV, do not do any scanning")
    .add_option(OPTION_DEVICE_BY_ID,   0, "device-by-id",   "VENDOR:PRODUCT", "Use device that matches VENDOR:PRODUCT (as returned by lsusb)")
    .add_newline()

    .add_text("Status Options: ")
    .add_option(OPTION_LED,     'l', "led",    "NUM", "set LED status, see --list-led-values (default: 0)")
    .add_option(OPTION_RUMBLE,  'r', "rumble", "L,R", "set the speed for both rumble motors [0-255] (default: 0,0)")
    .add_option(OPTION_QUIT,    'q', "quit",   "",    "only set led and rumble status then quit")
    .add_newline()

    .add_text("Configuration Options: ")
    .add_option(OPTION_DEADZONE,           0, "deadzone",         "INT",  "Threshold under which axis events are ignored (default: 0)")
    .add_option(OPTION_DEADZONE_TRIGGER,   0, "deadzone-trigger", "INT",  "Threshold under which trigger events are ignored (default: 0)")
    .add_option(OPTION_TRIGGER_AS_BUTTON,  0, "trigger-as-button", "",    "LT and RT send button instead of axis events")
    .add_option(OPTION_TRIGGER_AS_ZAXIS,   0, "trigger-as-zaxis", "",     "Combine LT and RT to form a zaxis instead")
    .add_option(OPTION_DPAD_AS_BUTTON,     0, "dpad-as-button",   "",     "DPad sends button instead of axis events")
    .add_option(OPTION_DPAD_ONLY,          0, "dpad-only",        "",     "Both sticks are ignored, only DPad sends out axis events")
    .add_option(OPTION_TYPE,               0, "type",             "TYPE", "Ignore autodetection and enforce controller type (xbox, xbox-mat, xbox360, xbox360-wireless, xbox360-guitar)")
    .add_option(OPTION_BUTTONMAP,         'b', "buttonmap",      "MAP",   "Remap the buttons as specified by MAP (example: B=A,X=A,Y=A)")
    .add_option(OPTION_AXISMAP,           'a', "axismap",        "MAP",   "Remap the axis as specified by MAP (example: -Y1=Y1,X1=X2)")
    .add_option(OPTION_NAME,               0, "name",             "DEVNAME", "Changes the descriptive name the device will have")
    .add_option(OPTION_UI_CLEAR,           0, "ui-clear",         "",     "Removes all existing uinput bindings")
    .add_option(OPTION_UI_BUTTONMAP,       0, "ui-buttonmap",     "MAP",  "Changes the uinput events send when hitting a button (example: X=BTN_Y,A=KEY_A)")
    .add_option(OPTION_UI_AXISMAP,         0, "ui-axismap",       "MAP",  "Changes the uinput events send when moving a axis (example: X1=ABS_X2)")
    .add_option(OPTION_SQUARE_AXIS,        0, "square-axis",      "",     "Cause the diagonals to be reported as (1,1) instead of (0.7, 0.7)")
    .add_option(OPTION_RELATIVE_AXIS,      0, "relative-axis",    "MAP",  "Make an axis emulate a joystick throttle (example: y2=64000)")
    .add_option(OPTION_AUTOFIRE,           0, "autofire",         "MAP",  "Cause the given buttons to act as autofire (example: A=250)")
    .add_option(OPTION_CALIBRARIOTION,     0, "calibration",      "MAP",  "Changes the calibration for the given axis (example: X2=-32768:0:32767)")
    .add_option(OPTION_FORCE_FEEDBACK,     0, "force-feedback",   "",     "Enable force feedback support")
    .add_option(OPTION_RUMBLE_GAIN,        0, "rumble-gain",      "NUM",  "Set relative rumble strength (default: 255)")
    .add_newline()

    .add_text("See README for more documentation and examples.")
    .add_text("Report bugs to Ingo Ruhnke <grumbel@gmx.de>");
}

void set_ui_button_map(ButtonEvent* ui_button_map, const std::string& str)
{
  std::string::size_type i = str.find_first_of('=');
  if (i == std::string::npos)
    {
      throw std::runtime_error("Couldn't convert string \"" + str + "\" to ui-button-mapping, '=' missing");
    }
  else
    {
      //std::cout << string2btn(str.substr(0, i)) << " -> " << str.substr(i+1, str.size()-i) << std::endl;

      XboxButton  btn   = string2btn(str.substr(0, i));
      ButtonEvent event = ButtonEvent::from_string(str.substr(i+1, str.size()-i));
      
      if (btn != XBOX_BTN_UNKNOWN)
        {
          ui_button_map[btn] = event;
        }
      else
        {
          throw std::runtime_error("Couldn't convert string \"" + str + "\" to ui-button-mapping, Xbox button name not valid");
        }
    }
}

void set_ui_axis_map(AxisEvent* ui_axis_map, const std::string& str)
{
  std::string::size_type i = str.find_first_of('=');
  if (i == std::string::npos)
    {
      throw std::runtime_error("Couldn't convert string \"" + str + "\" to ui-axis-mapping");
    }
  else
    {
      XboxAxis  axis  = string2axis(str.substr(0, i));
      AxisEvent event = AxisEvent::from_string(str.substr(i+1, str.size()-i));
            
      if (axis != XBOX_AXIS_UNKNOWN)
        {
          ui_axis_map[axis] = event;
        }
      else
        {
          throw std::runtime_error("Couldn't convert string \"" + str + "\" to ui-axis-mapping");
        }      
    }  
}

void
CommandLineOptions::parse_args(int argc, char** argv)
{  
  ArgParser::ParsedOptions parsed = argp.parse_args(argc, argv);

  for(ArgParser::ParsedOptions::const_iterator i = parsed.begin(); i != parsed.end(); ++i)
    {
      CommandLineOptions& opts = *this;
      const ArgParser::ParsedOption& opt = *i;

      switch (i->key)
        {
          case OPTION_HELP:
            opts.mode = PRINT_HELP;
            break;

          case OPTION_VERSION:
            opts.mode = PRINT_VERSION;
            break;
          
          case OPTION_VERBOSE:
            opts.verbose = true;
            break;

          case OPTION_QUIET:
            opts.quiet   = true;
            break;

          case OPTION_SILENT:
            opts.silent = true;
            break;

          case OPTION_DAEMON:
            opts.silent = true;
            opts.mode = RUN_DAEMON;
            break;

          case OPTION_TEST_RUMBLE:
            opts.rumble = true;
            break;

          case OPTION_RUMBLE:
            if (sscanf(opt.argument.c_str(), "%d,%d", &opts.rumble_l, &opts.rumble_r) == 2)
              {
                opts.rumble_l = std::max(0, std::min(255, opts.rumble_l));
                opts.rumble_r = std::max(0, std::min(255, opts.rumble_r));
              }
            else
              {
                RAISE_EXCEPTION(opt.option << " expected an argument in form INT,INT");
              }
            break;

          case OPTION_QUIT:
            opts.instant_exit = true;
            break;

          case OPTION_NO_UINPUT:
            opts.no_uinput = true;
            break;

          case OPTION_MIMIC_XPAD:
            //opts.mimic_xpad = true;
            break;

          case OPTION_TYPE:
            if (opt.argument == "xbox")
              {
                opts.gamepad_type = GAMEPAD_XBOX;
              }
            else if (opt.argument == "xbox-mat")
              {
                opts.gamepad_type = GAMEPAD_XBOX_MAT;
              }
            else if (opt.argument == "xbox360")
              {
                opts.gamepad_type = GAMEPAD_XBOX360;
              }
            else if (opt.argument == "xbox360-guitar")
              {
                opts.gamepad_type = GAMEPAD_XBOX360_GUITAR;
              }
            else if (opt.argument == "xbox360-wireless")
              {
                opts.gamepad_type = GAMEPAD_XBOX360_WIRELESS;
              }
            else if (opt.argument == "firestorm")
              {
                opts.gamepad_type = GAMEPAD_FIRESTORM;
              }
            else if (opt.argument == "firestorm-vsb")
              {
                opts.gamepad_type = GAMEPAD_FIRESTORM_VSB;
              }
            else if (opt.argument == "saitek-p2500")
              {
                opts.gamepad_type = GAMEPAD_SAITEK_P2500;
              }
            else
              {
                RAISE_EXCEPTION("unknown type: " << opt.argument << '\n'
                                << "Possible types are:\n"
                                << " * xbox\n"
                                << " * xbox-mat\n"
                                << " * xbox360\n"
                                << " * xbox360-guitar\n"
                                << " * xbox360-wireless\n"
                                << " * firestorm\n"
                                << " * firestorm-vsb\n"
                                << " * saitek-p2500\n");
              }
            break;

          case OPTION_FORCE_FEEDBACK:
            opts.uinput_config.force_feedback = true;
            break;

          case OPTION_RUMBLE_GAIN:
            opts.rumble_gain = to_number(255, opt.argument);
            break;

          case OPTION_BUTTONMAP:
            arg2vector(opt.argument, opts.button_map, &ButtonMapping::from_string);
            break;

          case OPTION_AXISMAP:
            arg2vector(opt.argument, opts.axis_map, &AxisMapping::from_string);
            break;
                    
          case OPTION_NAME:
            opts.uinput_config.device_name = opt.argument;
            break;

          case OPTION_UI_CLEAR:
            std::fill_n(opts.uinput_config.axis_map, (int)XBOX_AXIS_MAX, AxisEvent::invalid());
            std::fill_n(opts.uinput_config.btn_map,  (int)XBOX_BTN_MAX,  ButtonEvent::invalid());
            break;

          case OPTION_UI_AXISMAP:
            arg2apply(opt.argument, boost::bind(&set_ui_axis_map, opts.uinput_config.axis_map, _1));
            break;

          case OPTION_UI_BUTTONMAP:
            arg2apply(opt.argument, boost::bind(&set_ui_button_map, opts.uinput_config.btn_map, _1));
            break;

          case OPTION_ID:
            opts.controller_id = boost::lexical_cast<int>(opt.argument);
            break;

          case OPTION_WID:
            opts.wireless_id = boost::lexical_cast<int>(opt.argument);
            break;

          case OPTION_LED:
            if (opt.argument == "help")
              {
                opts.mode = PRINT_LED_HELP;
              }
            else
              {
                opts.led = boost::lexical_cast<int>(opt.argument);
              }
            break;
            
          case OPTION_DPAD_ONLY:
            if (opts.uinput_config.dpad_as_button)
              RAISE_EXCEPTION("Can't combine --dpad-as-button with --dpad-only");
            
            opts.uinput_config.dpad_only = true;
            break;
            
          case OPTION_DPAD_AS_BUTTON:
            if (opts.uinput_config.dpad_only)
              throw std::runtime_error("Can't combine --dpad-as-button with --dpad-only");

            opts.uinput_config.dpad_as_button = true;
            break;

          case OPTION_DEADZONE:
            opts.deadzone = to_number(32767, opt.argument);
            break;

          case OPTION_DEADZONE_TRIGGER:
            opts.deadzone_trigger = to_number(255, opt.argument);
            break;

          case OPTION_TRIGGER_AS_BUTTON:
            if (opts.uinput_config.trigger_as_zaxis)
              {
                RAISE_EXCEPTION("Can't combine --trigger-as-button and --trigger-as-zaxis");
              }
            else
              {
                opts.uinput_config.trigger_as_button = true;
              }
            break;

          case OPTION_AUTOFIRE:
            arg2vector(opt.argument, opts.autofire_map, &AutoFireMapping::from_string);
            break;

          case OPTION_CALIBRARIOTION:
            arg2vector(opt.argument, opts.calibration_map, &CalibrationMapping::from_string);
            break;

          case OPTION_RELATIVE_AXIS:
            arg2vector(opt.argument, opts.relative_axis_map, &RelativeAxisMapping::from_string);
            break;

          case OPTION_SQUARE_AXIS:
            opts.square_axis = true;
            break;

          case OPTION_TRIGGER_AS_ZAXIS:
            if (opts.uinput_config.trigger_as_button)
              {
                RAISE_EXCEPTION("Can't combine --trigger-as-button and --trigger-as-zaxis");
              }
            else
              {
                opts.uinput_config.trigger_as_zaxis = true;
              }
            break;

          case OPTION_HELP_LED:
            opts.mode = PRINT_LED_HELP;
            break;

          case OPTION_DEVICE_BY_ID:
            {
              unsigned int product_id;
              unsigned int vendor_id;
              if (sscanf(opt.argument.c_str(), "%x:%x", &vendor_id, &product_id) == 2)
                {
                  opts.vendor_id  = vendor_id;
                  opts.product_id = product_id;
                }
              else
                {
                  RAISE_EXCEPTION(opt.option << " expected an argument in form PRODUCT:VENDOR (i.e. 046d:c626)");
                }
              break;
            }

          case OPTION_DEVICE_BY_PATH:
            if (sscanf(opt.argument.c_str(), "%3s:%3s", opts.busid, opts.devid) != 2)
              {  
                RAISE_EXCEPTION(opt.option << " expected an argument in form BUS:DEV (i.e. 006:003)");
              }
            break;

          case OPTION_LIST_SUPPORTED_DEVICES:
            opts.mode = RUN_LIST_SUPPORTED_DEVICES;
            break;

          case OPTION_LIST_CONTROLLER:
            opts.mode = RUN_LIST_CONTROLLER;
            break;

          case OPTION_HELP_DEVICES:
            opts.mode = PRINT_HELP_DEVICES;
            break;

          case ArgParser::REST_ARG:
            RAISE_EXCEPTION("unknown command line option: " << opt.argument);
            break;

          default:
            RAISE_EXCEPTION("unknown command line option: " << opt.option);
            break;
        }
    }
}

void
CommandLineOptions::print_help() const
{
  argp.print_help(std::cout);
}

void
CommandLineOptions::print_led_help() const
{
  std::cout << 
    "Possible values for '--led VALUE' are:\n\n"
    "   0: off\n"
    "   1: all blinking\n"
    "   2: 1/top-left blink, then on\n"
    "   3: 2/top-right blink, then on\n"
    "   4: 3/bottom-left blink, then on\n"
    "   5: 4/bottom-right blink, then on\n"
    "   6: 1/top-left on\n"
    "   7: 2/top-right on\n"
    "   8: 3/bottom-left on\n"
    "   9: 4/bottom-right on\n"
    "  10: rotate\n"
    "  11: blink\n"
    "  12: blink slower\n"
    "  13: rotate with two lights\n"
    "  14: blink\n"
    "  15: blink once\n"
            << std::endl;
}
  
void
CommandLineOptions::print_version() const
{
  std::cout
    << "xboxdrv 0.4.8\n"
    << "Copyright (C) 2008 Ingo Ruhnke <grumbel@gmx.de>\n"
    << "License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>\n"
    << "This is free software: you are free to change and redistribute it.\n"
    << "There is NO WARRANTY, to the extent permitted by law."
    << std::endl;
}
                
/* EOF */
